/*Vesola's Bot abilites
	by DirkTheHusky
Edit from Dirk's Bot abilites
	by Lach
Further optimizations from CobaltBW's bot code.
*/

local debug = false -- Set to true to output the botticcmd timeline
local dprint = function(str)
	return debug and print(str)
end

-- Vesola AI (ThinkState controller)
local setThink = function(ai, str, angle)
	str = $ or 'default'
	assert(type(str) == 'string', 'setThink argument #2 must be a string! Got argument "'..tostring(str)..'"')
	dprint('setting thinkstate from '..tostring(ai.thinkstate..' to '..str))
	ai.thinkstate = str
	ai.thinktime = 0
	ai.catchup_tics = 0
	if angle != nil
		ai.thinkangle = angle
	end
end

-- ThinkHelp radius
local thinkHelpRadius = function(bmo, pmo, radiusonly)
	return (radiusonly or (bmo.player.speed == 0 and P_IsObjectOnGround(bmo) and bmo.player.pflags & (PF_STARTDASH|PF_SPINNING) == 0) )
		and P_AproxDistance(bmo.z - pmo.z, P_AproxDistance(bmo.x - pmo.x, bmo.y - pmo.y)) <= bmo.radius*2
end

-- Look at player
local lookAtPlayer = function(bot, player)
	bot.cmd.angleturn = R_PointToAngle2(bot.mo.x, bot.mo.y, player.mo.x, player.mo.y) >> 16
end

-- Vesola AI (thinkstate functions)
local aiFunc = {
	 -- Standard follow code. This code also gets the AI to the alternative thinkstates.
	default = function(bot, cmd, ai, player)
		local mo = player.mo
		if P_IsObjectOnGround(bot.mo)
		and thinkHelpRadius(bot.mo, mo)
			-- Set thinkhelp state
			setThink(ai, 'thinkhelp')
			return true
		end
		if (ai.catchup_tics >= 70)
			setThink(ai, 'catchup')
			return true
		end
		-- Prevent AI Spin ability spam and only call them when we actually need them.
		if player.pflags & PF_SPINNING
			cmd.buttons = $ & ~BT_SPIN
		end
		-- Fallthrough to vanilla AI
		return false
	end,

	-- Thinking of helping bot leader
	thinkhelp = function(bot, cmd, ai, player)
		local mo = player.mo
		
		-- Not in radius
		if not thinkHelpRadius(bot.mo, mo)
			setThink(ai) -- Reset to default thinkstate
			return false
		end
		-- Determine if we're in a position where Vesola should use her jump ability
		if player.cmd.buttons & BT_JUMP
		and not (player.pflags & PF_THOKKED)
		and player.pflags & PF_JUMPED
		and P_MobjFlip(mo) * mo.momz > 0
		and P_IsObjectOnGround(bot.mo)
			setThink(ai, 'ability', mo.angle) -- Setting the angle to force the ai to look forwards
			-- Set inputs
			cmd.buttons = $ | BT_JUMP
			-- This is a bit of a cheat, but it gets the job done I guess
			return true
		end
		if player.cmd.buttons & BT_SPIN
			setThink(ai, 'ability2')
			bot.mo.angle = mo.angle
			cmd.buttons = $ | BT_SPIN
			return true
		end
		return true
	end,
	
	ability = function(bot, cmd, ai, player)
		local mo = player.mo
		dprint('jump')
		if bot.beeflight -- We entered the petal twirl
			setThink(ai, 'ability_twirling') -- Spinning
			return true
		end
		if not bot.beeflight
			if abs(bot.mo.momz) > 4*mo.scale/(mo.eflags & MFE_UNDERWATER and 3 or 1) -- If we're going fast enough up or down, press jump. The small gap inbetween means we will be releasing jump before pressing it again.
				cmd.buttons = $ | BT_JUMP
				return true
			end
		end
	end,
	
	ability_twirling = function(bot, cmd, ai, player)
		local mo = player.mo
		
		bot.mo.angle = cmd.angleturn << 16 -- Necessary hack, apparently
		bot.drawangle = bot.mo.angle -- Also a hack
		
		local pc = player.precmd
		-- Leader takes control
		cmd.angleturn = pc.angleturn - pc.sidemove * (ANG1 >> 16) -- Copies player angleturn.
		cmd.forwardmove = pc.forwardmove -- Copies player forwardmove
		cmd.sidemove = pc.sidemove -- Copies player forwardmove
		if player.powers[pw_carry] == CR_PLAYER
			if not (player.cmd.buttons & BT_SPIN)
				cmd.buttons = $ | BT_JUMP
				dprint('flying up')
			else
				cmd.buttons = $ & ~BT_JUMP
				dprint('flying down')
			end
			return true
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai) -- Reset to default thinkstate
		end
	end,
	
	-- We're planting the flower in.
	ability2 = function(bot, cmd, ai, player)
		local mo = player.mo
		if ai.thinktime & 2
			dprint('press spin')
			cmd.buttons = $ | BT_SPIN
			cmd.buttons = $ &~ BT_JUMP
		else
			dprint('release spin')
			cmd.buttons = $ &~ BT_SPIN
			cmd.buttons = $ &~ BT_JUMP
			setThink(ai) -- Reset to default thinkstate
		end
		return true
	end,
	
	catchup = function(bot, cmd, ai, player)
		local mo = player.mo
		
		dprint('catching up')
		
		lookAtPlayer(bot, player)
		
		local flip = P_MobjFlip(bot.mo)
		local zdist = flip * (mo.z - bot.mo.z)
		
		if zdist > 0
			cmd.buttons = $ | BT_JUMP
			setThink(ai, 'catchjump')
			return true
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
	end,
	
	catchjump = function(bot, cmd, ai, player)
		local mo = player.mo
		dprint('jump')
		
		lookAtPlayer(bot, player)
		cmd.forwardmove = 50
		
		if bot.beeflight -- We entered the petal twirl
			setThink(ai, 'catch_twirling') -- Spinning
			return true
		end
		if not bot.beeflight
			if abs(bot.mo.momz) > 4*mo.scale/(mo.eflags & MFE_UNDERWATER and 3 or 1) -- If we're going fast enough up or down, press jump. The small gap inbetween means we will be releasing jump before pressing it again.
				cmd.buttons = $ | BT_JUMP
				return true
			end
		end
	end,
	
	catch_twirling = function(bot, cmd, ai, player)
		local mo = player.mo
		
		lookAtPlayer(bot, player)
		cmd.forwardmove = 50
		
		if bot.beeflight
			if P_MobjFlip(bot.mo)*bot.mo.z > P_MobjFlip(bot.mo)*mo.z
				cmd.buttons = $ & ~BT_JUMP
				dprint('flying up')
			else
				cmd.buttons = $ | BT_JUMP
				dprint('falling')
			end
		else
			setThink(ai) -- Reset to default thinkstate
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai) -- Reset to default thinkstate
		end
		return true
	end
}

-- This hook sets player.precmd (necessary for surf controls)
addHook('PreThinkFrame',do
	for player in players.iterate do
		player.precmd = $ or {}
		local cmd = player.cmd
		local pc = player.precmd
		pc.buttons = cmd.buttons
		pc.forwardmove = cmd.forwardmove
		pc.sidemove = cmd.sidemove
		pc.angleturn = cmd.angleturn
		pc.aiming = cmd.aiming
	end
end)

local warning = false
-- Vesola AI (Master level)
addHook("BotTiccmd", function(bot,cmd)
	if paused or gamestate != GS_LEVEL
	or bot.mo == nil -- Must be playing
	or skins[bot.skin].name != 'vesola' -- Vesola only
	or not (bot.botleader and bot.botleader.mo) -- Must have a botleader
	or bot.botleader.exiting -- Must not be the end of the level
	or P_PlayerInPain(bot) or bot.playerstate != PST_LIVE -- Hurt and dead bots can't think
		bot.vesolaai = nil -- Unset Vesola AI vars
		return
	end	

	-- Initialize Vesola AI vars
	bot.vesolaai = $ or {
		thinkstate = 'default',
		thinktime = -1,
		catchup_tics = 0,
		thinkangle = angle,
		errorstate = false,
	}
	local ai = bot.vesolaai
	
	if ai.errorstate and warning == false -- The script should never be broken past this point. If it is, this will print a message to warn the user
		warning = true
		print('\x82\ Please check console/logs for errors associated with Vesola AI.')
		debug = false -- Disable debug print messages past this point.
	end
	
	-- Update Vars
	ai.thinkstate = $ or 'default' -- This can never be nil
	ai.thinktime = $+1
	ai.errorstate = true -- The end of the function needs to be reached to reset this to false.
	
	local mo = bot.botleader.mo
	local flip = P_MobjFlip(bot.mo)
	local dist = P_AproxDistance(mo.x - bot.mo.x, mo.y - bot.mo.y)
	local zdist = flip * (mo.z - bot.mo.z)
	local bmom = P_AproxDistance(bot.mo.momx, bot.mo.momy)
	local followmax = 128 * 8 * bot.mo.scale -- Max follow distance before AI begins to enter catchup state
	local followthres = 92 * bot.mo.scale -- Distance that AI will try to reach
	local comfortheight = 96 * bot.mo.scale
	local stalled = (bmom < bot.mo.scale >> 1) and dist > followthres -- Helps to see if the AI is having trouble catching up
	
	if (dist > followmax or zdist > comfortheight or stalled)	
		ai.catchup_tics = min($+2, 70)
	else
		ai.catchup_tics = max($-1, 0)
	end

	dprint('\x81\ vesola bot\x80 thinkstate '..tostring(ai.thinkstate)..' (time: '..tostring(ai.thinktime)..')')
		
	-- Perform thinkstate's function
	local thinkFunc = aiFunc[ai.thinkstate]
	assert(type(thinkFunc) == "function", "No think function for Vesola AI thinkstate '"..tostring(ai.thinkstate).."' (got "..tostring(thinkFunc)..")")
	local returnVal = thinkFunc(bot, cmd, ai, bot.botleader) -- The return value will control whether we will fallthrough to SRB2's native TailsAI code.
	
	local player = bot.botleader
	local pmo = player.mo		

	-- Helper indicator
	if bot.mo.vesolaind
		-- Display indicator
		if ai.thinkstate == 'thinkhelp'
			dprint('vesola indicator (draw)')
			bot.mo.vesolaind.flags2 = $ & ~MF2_DONTDRAW
		else -- Do not display indicator
			dprint("vesola indicator (don't draw)")
			bot.mo.vesolaind.flags2 = $|MF2_DONTDRAW
		end
	else
		dprint('initialize vesola indicator')
		local ind = P_SpawnMobj(bot.mo.x, bot.mo.y, bot.mo.z, MT_OVERLAY)
		ind.target = bot.mo
		ind.state = S_FLIGHTINDICATOR
		ind.flags2 = $|MF2_DONTDRAW
		bot.mo.vesolaind = ind
	end
	
	-- Post-update variables
	ai.errorstate = false
		
	dprint('fallthrough: '..tostring(not returnVal))
	return returnVal -- Returns the value received from vesolaai's thinkstate function
end)