/*
Whirl's Bot abilites
	by SMS Alfredo
Edit from Dirk's Bot abilites
	by Lach
Further optimizations from CobaltBW's bot code.
*/
local debug = false -- Set to true to output the botticcmd timeline
local dprint = function(str)
	return debug and print(str)
end

-- Whirl AI (ThinkState controller)
local setThink = function(ai, str, angle)
	str = $ or 'default'
	assert(type(str) == 'string', 'setThink argument #2 must be a string! Got argument "'..tostring(str)..'"')
	dprint('setting thinkstate from '..tostring(ai.thinkstate..' to '..str))
	ai.thinkstate = str
	ai.thinktime = 0
	ai.catchup_tics = 0
	if angle != nil
		ai.thinkangle = angle
	end
end

-- ThinkHelp radius
local thinkHelpRadius = function(bmo, pmo, radiusonly)
	return (radiusonly or (bmo.player.speed == 0 and P_IsObjectOnGround(bmo) and bmo.player.pflags & (PF_STARTDASH|PF_SPINNING) == 0) )
		and P_AproxDistance(bmo.z - pmo.z, P_AproxDistance(bmo.x - pmo.x, bmo.y - pmo.y)) <= bmo.radius*2
end

-- If the bot is in the radius
local botInRadius = function(bmo, pmo, radiusonly)
	return (radiusonly or bmo.player.whirltornadojump)
		and P_AproxDistance(0, P_AproxDistance(bmo.x - pmo.x, bmo.y - pmo.y)) <= bmo.radius*16
end

-- Look at player
local lookAtPlayer = function(bot, player)
	bot.cmd.angleturn = R_PointToAngle2(bot.mo.x, bot.mo.y, player.mo.x, player.mo.y) >> 16
end

-- A lot of death pit checker functions. Thanks CobaltBW
------------------------------------------------

local GetFloorSector = function(mo, flip, allowfof)
	flip = $ or P_MobjFlip(mo)
	if allowfof == false
		return mo.subsector.sector
	else
		return flip == 1 and mo.floorrover and mo.floorrover.sector
			or flip == -1 and mo.ceilingrover and mo.ceilingrover.sector
			or mo.subsector.sector
	end
end

local GetFloorSpecial = function(sector, specialtype, flip)
	specialtype = $ or 1
	local spec = GetSecSpecial(sector.special, specialtype or 1)
	local n, line = -1
	 -- Do floor check
	if flip
	and specialtype == 1 and spec != 8 and spec != 10 and spec != 12
		repeat -- Iterate through lines
			n = P_FindSpecialLineFromTag(8, sector.tag, $)
			line = lines[n]
			if not(line) continue end
			if flip == 1 and line.flags & (1<<6) -- Floor not affected
				return 0
			elseif flip == -1 and line.flags & ((1<<6)|(1<<9)) -- Ceiling affected
				return spec
			end
		until line == nil
		-- Fallthrough for flipped gravity
		if flip == -1
			return 0
		end
	end
	-- Fallthrough for normal gravity
	return spec
end

local GetDangerCrumble = function(rover)
	if userdataType(rover) == 'mobj_t'
		rover = P_MobjFlip($) == 1 and $.floorrover or $.ceilingrover
	end
	return rover and rover.flags & (FF_BLOCKPLAYER|FF_CRUMBLE) == FF_BLOCKPLAYER|FF_CRUMBLE and rover.flags & (FF_FLOATBOB|FF_NORETURN) != FF_FLOATBOB|FF_NORETURN
end
 
local GetDangerFloor = function(mo)
	local flip = P_MobjFlip(mo)	
	local sector = 
		flip == 1 and mo.floorrover
		or flip == -1 and mo.ceilingrover
	local fof = false
	//Determine if FOF is a crumbling platform
	if sector
		local crumble = GetDangerCrumble(sector)
		if not(crumble)
			sector = $.sector
			fof = true
		else
			sector = nil
		end
	end
	if not(sector) //Use floor instead
		sector = GetFloorSector(mo, flip, false)
	end
-- 	Nf.TrueFalseGhost(mo, fof)
	local spec = GetFloorSpecial(sector, 1, flip)
	return (spec and spec <= 12)
end
------------------------------------------------

-- Whirl AI (thinkstate functions)
local aiFunc = {
	 -- Standard follow code. This code also gets the AI to the alternative thinkstates.
	default = function(bot, cmd, ai, player)
		local mo = player.mo		
		
		if P_IsObjectOnGround(bot.mo)
		and thinkHelpRadius(bot.mo, mo)
			-- Set thinkhelp state
			setThink(ai, 'thinkhelp')
			return true
		end
		
		if (ai.catchup_tics >= 70)
			setThink(ai, 'catchup')
			return true
		end
		-- Fallthrough to vanilla AI
		return false
	end,

	-- Thinking of helping bot leader
	thinkhelp = function(bot, cmd, ai, player)
		local mo = player.mo
				
		-- Not in radius
		if not thinkHelpRadius(bot.mo, mo)
			setThink(ai) -- Reset to default thinkstate
			return false
		end
		-- Determine if we're in a position where Whirl should use his jump ability
		if player.cmd.buttons & BT_JUMP
		and not (player.pflags & PF_THOKKED)
		and player.pflags & PF_JUMPED
		and P_MobjFlip(mo) * mo.momz > 0
		and P_IsObjectOnGround(bot.mo)
			setThink(ai, 'ability', mo.angle) -- Setting the angle to force the ai to look forwards
			-- Set inputs
			cmd.buttons = $ | BT_JUMP
			cmd.angleturn = ai.thinkangle >> 16
			-- This is a bit of a cheat, but it gets the job done I guess
			bot.mo.angle = cmd.angleturn << 16
			bot.jumpfactor = FRACUNIT
			return true
		end
		lookAtPlayer(bot, player)
		return true
	end,
	
	-- Bot falling behind?
	catchup = function(bot, cmd, ai, player)
		local mo = player.mo
		
		dprint('catching up')
		
		local flip = P_MobjFlip(bot.mo)
		local zdist = flip * (mo.z - bot.mo.z)
		
		if zdist > 0
			cmd.buttons = $ | BT_JUMP
			setThink(ai, 'catchjump')
			return true
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
	end,
	
	catchjump = function(bot, cmd, ai, player)
		local mo = player.mo
		dprint('jump')
		
		if bot.whirltornadojump -- We jumped
			setThink(ai, 'aircatching')
			bot.mo.momz = $*2
			return true
		end
		if not bot.whirltornadojump
			if abs(bot.mo.momz) > 4*mo.scale/(mo.eflags & MFE_UNDERWATER and 3 or 1) -- If we're going fast enough up or down, press jump. The small gap inbetween means we will be releasing jump before pressing it again.
				cmd.buttons = $ | BT_JUMP
			end
			return true
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
	end,
	
	aircatching = function(bot, cmd, ai, player)
		local mo = player.mo
		
		local  dist = P_AproxDistance(mo.x - mo.x, bot.mo.y - bot.mo.y)
		local  followmax = 128 * 8 * bot.mo.scale -- Max follow distance before AI begins to enter catchup state
		local  followmin = 32 * bot.mo.scale
		
		if botInRadius(bot.mo, mo)
		and not ((P_MobjFlip(bot.mo) * bot.mo.momz) > 0)
			setThink(ai, 'ability_drilling')
		end

		cmd.buttons = $ & ~BT_JUMP -- You're not pressing jump
		lookAtPlayer(bot,player) -- Bot is following the player
		cmd.forwardmove = 50 
		-- If the bot is on the ground, reset it!
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
		return true
	end,
	
	-- Jump code, let's make whirl fly high!
	ability = function(bot, cmd, ai, player)
		local mo = player.mo
		dprint('jump')
		if bot.whirltornadojump -- We jumped
			setThink(ai, 'ability_aircontrol')
			bot.mo.momz = $*2
			return true
		end
		if not bot.whirltornadojump
			if abs(bot.mo.momz) > 4*mo.scale/(mo.eflags & MFE_UNDERWATER and 3 or 1) -- If we're going fast enough up or down, press jump. The small gap inbetween means we will be releasing jump before pressing it again.
				cmd.buttons = $ | BT_JUMP
			end
			return true
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
	end,
	
	-- Did we successfully grab on to the bot? let's get it to think for its air control!
	ability_aircontrol = function(bot, cmd, ai, player)
		local mo = player.mo
		
		bot.mo.angle = cmd.angleturn << 16 -- Necessary hack, apparently
		bot.drawangle = bot.mo.angle -- Also a hack
		
		local pc = player.precmd
		-- Leader takes control
		cmd.angleturn = pc.angleturn - pc.sidemove * (ANG1 >> 16) -- Copies player angleturn.
		cmd.forwardmove = pc.forwardmove -- Copies player forwardmove
		cmd.sidemove = pc.sidemove -- Copies player forwardmove
		
		if not bot.whirltornadojump
			setThink(ai) -- Reset thinkstate
			bot.jumpfactor = skins[bot.mo.skin].jumpfactor
		else
			if player.powers[pw_carry] == CR_PLAYER -- Are we latching on?
				cmd.buttons = $ & ~BT_JUMP -- Let go of the jump button if we latched on.
				dprint('latching')
			end
		end
		
		if bot.whirltornadojump
			if not player.powers[pw_carry]
				setThink(ai, 'ability_playerlook')
				cmd.buttons = $ & ~BT_JUMP -- You're not pressing jump when I jump off.
			end
		end
				
		if player.cmd.buttons & BT_SPIN 
		or P_IsObjectOnGround(mo) -- If the leader is already on the ground.
			setThink(ai, 'ability_drilling') -- Drill.
			dprint('drilling')
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
		return true
	end,
	
	ability_playerlook = function(bot, cmd, ai, player)
		local mo = player.mo
		
		dprint('player looking')
		cmd.buttons = $ & ~BT_JUMP -- You still not pressing jump
		lookAtPlayer(bot,player) -- Bot is following the player
		cmd.forwardmove = 50 
		if P_IsObjectOnGround(mo)
		and not GetDangerFloor(bot.mo)
			setThink(ai, 'ability_drilling') -- Drill.
			dprint('drilling')
		end
		-- If the bot is on the ground, reset it!
		if P_IsObjectOnGround(bot.mo)
			setThink(ai)
		end
		-- If we latch back on, we control again.
		if bot.whirltornadojump
		and player.powers[pw_carry]
			setThink(ai, 'ability_aircontrol')
			return true
		end
		return true
	end,
	
	ability_drilling = function(bot, cmd, ai, player)
		if bot.whirltornadojump
			cmd.buttons = $ | BT_SPIN
		end
		if P_IsObjectOnGround(bot.mo)
			setThink(ai) -- Bot has landed, back to AI thinking
		end
		return true
	end
}

-- This hook sets player.precmd (necessary for air control)
addHook('PreThinkFrame',do
	for player in players.iterate do
		player.precmd = $ or {}
		local cmd = player.cmd
		local pc = player.precmd
		pc.buttons = cmd.buttons
		pc.forwardmove = cmd.forwardmove
		pc.sidemove = cmd.sidemove
		pc.angleturn = cmd.angleturn
		pc.aiming = cmd.aiming
	end
end)

local warning = false
-- Whirl AI (Master level)
addHook("BotTiccmd", function(bot,cmd)
	if paused or gamestate != GS_LEVEL
	or bot.mo == nil -- Must be playing
	or skins[bot.skin].name != 'whirl' -- Whirl only
	or not (bot.botleader and bot.botleader.mo) -- Must have a botleader
	or bot.botleader.exiting -- Must not be the end of the level
	or P_PlayerInPain(bot) or bot.playerstate != PST_LIVE -- Hurt and dead bots can't think
		bot.whirlai = nil -- Unset Whirl AI vars
		return
	end
	
	-- Initialize Whirl AI vars
	bot.whirlai = $ or {
		thinkstate = 'default',
		thinktime = -1,
		catchup_tics = 0,
		thinkangle = angle,
		errorstate = false,
	}
	local ai = bot.whirlai
	
	local mo = bot.botleader.mo
	local flip = P_MobjFlip(bot.mo)
	local dist = P_AproxDistance(mo.x - bot.mo.x, mo.y - bot.mo.y)
	local zdist = flip * (mo.z - bot.mo.z)
	local bmom = P_AproxDistance(bot.mo.momx, bot.mo.momy)
	local followmax = 128 * 8 * bot.mo.scale -- Max follow distance before AI begins to enter catchup state
	local followthres = 92 * bot.mo.scale -- Distance that AI will try to reach
	local comfortheight = 96 * bot.mo.scale
	local stalled = (bmom < bot.mo.scale >> 1) and dist > followthres -- Helps to see if the AI is having trouble catching up
	
	if ai.errorstate and warning == false -- The script should never be broken past this point. If it is, this will print a message to warn the user
		warning = true
		print('\x82\ Please check console/logs for errors associated with Whirl AI.')
		debug = false -- Disable debug print messages past this point.
	end
	
	-- Update Vars
	ai.thinkstate = $ or 'default' -- This can never be nil
	ai.thinktime = $+1
	ai.errorstate = true -- The end of the function needs to be reached to reset this to false.
	
	if (dist > followmax or zdist > comfortheight or stalled)	
		ai.catchup_tics = min($+2, 70)
	else
		ai.catchup_tics = max($-1, 0)
	end

	dprint('\x81\ whirl bot\x80 thinkstate '..tostring(ai.thinkstate)..' (time: '..tostring(ai.thinktime)..')')
		
	-- Perform thinkstate's function
	local thinkFunc = aiFunc[ai.thinkstate]
	assert(type(thinkFunc) == "function", "No think function for Whirl AI thinkstate '"..tostring(ai.thinkstate).."' (got "..tostring(thinkFunc)..")")
	local returnVal = thinkFunc(bot, cmd, ai, bot.botleader) -- The return value will control whether we will fallthrough to SRB2's native TailsAI code.
	
	local player = bot.botleader
	local pmo = player.mo		


	-- Helper indicator
	if bot.mo.whirlind
		-- Display indicator
		if ai.thinkstate == 'thinkhelp'
			dprint('whirl indicator (draw)')
			bot.mo.whirlind.flags2 = $ & ~MF2_DONTDRAW
		else -- Do not display indicator
			dprint("whirl indicator (don't draw)")
			bot.mo.whirlind.flags2 = $|MF2_DONTDRAW
		end
	else
		dprint('initialize whirl indicator')
		local ind = P_SpawnMobj(bot.mo.x, bot.mo.y, bot.mo.z, MT_OVERLAY)
		ind.target = bot.mo
		ind.state = S_FLIGHTINDICATOR
		ind.flags2 = $|MF2_DONTDRAW
		bot.mo.whirlind = ind
	end	
	
	-- Post-update variables
	ai.errorstate = false
		
	dprint('fallthrough: '..tostring(not returnVal))
	return returnVal -- Returns the value received from whirlai's thinkstate function
end)